import os
from typing import Any, Callable, Dict, List, NamedTuple, Optional

from django.conf import settings
from django.http import HttpRequest
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

from .renderer import render_jsx_to_string
from .serialization import create_schema, serialize


class JSX(BaseEngine):
    # Name of the subdirectory containing the templates for this engine
    # inside an installed application.
    # app_dirname = 'foobar'

    def __init__(self, params: Any) -> None:
        # params = params.copy()
        options = params.pop("OPTIONS").copy()

        self.context_processors = options.pop("context_processors", [])
        super().__init__(params)

        # self.engine = foobar.Engine(**options)

    def from_string(self, template_code: str) -> Any:
        raise TemplateSyntaxError("Unsupported with JSX")

    def get_template(self, template_name: str) -> "JSXTemplate":  # type: ignore[override]
        adapter = self.template_adapters.get(template_name)

        if adapter is not None:
            return AdapterTemplate(adapter, self)

        if template_name.endswith(".tsx") or template_name.endswith(".jsx"):
            if os.path.isfile(
                os.path.join(settings.BASE_DIR, "client/templates", template_name)
            ):
                return JSXTemplate(template_name, self)

        raise TemplateDoesNotExist([], backend=self)  # type: ignore[arg-type]

    @cached_property
    def template_adapters(self) -> Dict[str, NamedTuple]:
        adapters = {}

        [
            adapters.update(import_string(path))
            for path in getattr(settings, "REACTIVATED_ADAPTERS", [])
        ]

        return adapters

    @cached_property
    def template_context_processors(self) -> List[Callable[[HttpRequest], Any]]:
        return [import_string(path) for path in self.context_processors]


class JSXTemplate:
    def __init__(self, jsx_template_name: str, backend: JSX) -> None:
        self.jsx_template_name = jsx_template_name
        self.backend = backend

    def render(self, context: Any = None, request: Optional[HttpRequest] = None) -> str:
        template_name = self.jsx_template_name.replace(".tsx", "").replace(".jsx", "")
        from .serialization.context_processors import (
            create_context_processor_type,
            BaseContext,
        )

        props = context or {}
        react_context = BaseContext(template_name=template_name)._asdict()

        if request is not None:
            for context_processor in self.backend.template_context_processors:
                react_context.update(context_processor(request))

            serialized_context = serialize(
                react_context,
                create_schema(
                    create_context_processor_type(self.backend.context_processors), {}
                ),
            )

            return render_jsx_to_string(request, serialized_context, props)

        assert (
            False
        ), "At this time, only templates with the request object can be rendered with reactivated"


class AdapterTemplate(JSXTemplate):
    def __init__(self, adapter: Any, backend: Any) -> None:
        self.adapter = adapter
        super().__init__(f"{self.adapter.__name__}.tsx", backend)

    def render(self, context: Any = None, request: Optional[HttpRequest] = None) -> str:
        to_be_serialized = self.adapter(**context)
        jsx_context = to_be_serialized.get_serialized()
        return super().render(context=jsx_context, request=request)



import atexit
import logging
import re
import subprocess
import sys
from typing import Any, List, Optional

import requests
import simplejson
from django.conf import settings
from django.http import HttpRequest
from django.template.defaultfilters import escape

renderer_process_port = None
logger = logging.getLogger("django.server")


def wait_and_get_port() -> Optional[int]:
    global renderer_process_port

    if renderer_process_port is not None:
        return renderer_process_port

    renderer_command = (
        ["node", "node_modules/reactivated/renderer.js",]
        if settings.DEBUG is False
        else [
            "node_modules/.bin/babel-node",
            "--extensions",
            ".ts,.tsx",
            "node_modules/reactivated/renderer.js",
        ]
    )

    logger.info("Starting render process")

    process = subprocess.Popen(
        renderer_command, encoding="utf-8", stdout=subprocess.PIPE,
    )

    def cleanup() -> None:
        # Pytest has issues with this, see https://github.com/pytest-dev/pytest/issues/5502
        # We can't use the env variable PYTEST_CURRENT_TEST because this happens
        # after running all tests and closing the session.
        # See: https://stackoverflow.com/questions/25188119/test-if-code-is-executed-from-within-a-py-test-session
        if "pytest" not in sys.modules:
            logger.info("Cleaning up renderer process")
        process.terminate()

    atexit.register(cleanup)

    output = ""

    for c in iter(lambda: process.stdout.read(1), b""):  # type: ignore[union-attr]
        output += c

        if match := re.match(r"RENDERER:([\d]+):LISTENING", output):
            renderer_process_port = int(match.group(1))
            return renderer_process_port
    assert False, "Could not bind to renderer"


def get_accept_list(request: HttpRequest) -> List[str]:
    """
    Given the incoming request, return a tokenized list of media
    type strings.

    From https://github.com/encode/django-rest-framework/blob/master/rest_framework/negotiation.py
    """
    header = request.META.get("HTTP_ACCEPT", "*/*")
    return [token.strip() for token in header.split(",")]


def should_respond_with_json(request: HttpRequest) -> bool:
    accepts = get_accept_list(request)

    return request.GET.get("format", None) == "json" or any(
        ["application/json" in content_type for content_type in accepts]
    )


def render_jsx_to_string(request: HttpRequest, context: Any, props: Any) -> str:
    respond_with_json = should_respond_with_json(request)

    payload = {"context": context, "props": props}
    data = simplejson.dumps(payload, indent=4)
    headers = {"Content-Type": "application/json"}

    if "debug" in request.GET:
        return f"<html><body><h1>Debug response</h1><pre>{escape(data)}</pre></body></html>"
    elif (
        respond_with_json or "raw" in request.GET or settings.REACTIVATED_SERVER is None
    ):
        request._is_reactivated_response = True  # type: ignore[attr-defined]
        return data

    renderer_port = wait_and_get_port()

    response = requests.post(
        f"http://localhost:{renderer_port}", headers=headers, data=data
    )

    if response.status_code == 200:
        return response.text
    else:
        raise Exception(response.json()["stack"])
